/*
  untested, not working (incomplete set of funcs for a start)

  mpr_crt__urandomm not right
*/



/* Chinese Remainder Theorem based modular arithmetic.

   Copyright 2002, 2007 Kevin Ryde

   This file is part of MPM.

   MPM is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2, or (at your option) any later version.

   MPM is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along
   with MPM; see the file COPYING.  If not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
   USA.  */

#include <stdio.h>
#include <stdlib.h>
#include "gmp.h"
#include "gmp-impl.h"
#include "longlong.h"

#include "mpm.h"
#include "mpm-impl.h"
#include "redc1.h"


/* This is for the special case of a modulus which can be factorized into
   factors each of which fits within a limb (and nails), and all of which
   pairwise have no common factor (ie. are co-prime).  Such factors don't
   have to be primes, just arranged so that each prime appears in just one
   factor.  Under the Chinese remainder theorem calculations can be made
   modulo each factor, then combined at the end for a final result.

   The modulus is taken by mpm_crt_init() as a limb-factor array, and is
   also kept multiplied up to a bignum internally.  Residues are limb arrays
   with each element kept within the range 0 <= x < factor.  The advantage
   of this style is that there's no carries between limbs, so it vectorizes
   on suitable CPUs.

   ENHANCE-ME: Could use REDC as per redc1.h if the modulus is odd.

   ENHANCE-ME: Interleaving f and finv elements in fp would best suit scalar
   chips, but separate fp and fpinv might suit cray better.  */


/* array of limb moduli */
#define MFP(m)  ((m)->_mp_inv)

#if WANT_ASSERT
#define ASSERT_MPR_CRT(r,m)                     \
  do {                                          \
    mp_size_t  __i;                             \
    for (__i = 0; __i < MRSIZE(m); __i++)       \
      ASSERT (RPTR(r)[__i] < MFP(m)[__i]);      \
  } while (0)
#else
#define ASSERT_MPR_CRT(r,m)  do { } while (0)
#endif


static void
mpr_crt__add (mpr_ptr w, mpr_srcptr x, mpr_srcptr y, mpm_srcptr m)
{
  mp_srcptr  fp, xp, yp;
  mp_size_t  fn, i;
  mp_ptr     wp;

  ASSERT_MPR_CRT (x, m);
  ASSERT_MPR_CRT (y, m);

  fp = MFP(m);
  fn = MRSIZE(m);
  wp = RPTR (w);
  xp = RPTR (x);
  yp = RPTR (y);

  CRAY_Pragma ("_CRI ivdep");
  for (i = 0; i < fn; i++)
    ADDMOD_1 (wp[i], xp[i], yp[i], fp[i]);

  ASSERT_MPR_CRT (w, m);
}

/* ENHANCE-ME: Pre-calculate some flimb inverses for y%flimb.
   ENHANCE-ME: Note the smallest flimb, so if y < min(flimb) a loop without
   y%flimb can be used (good for small y).  */
static void
mpr_crt__add_ui (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_srcptr  xp, fp;
  mp_ptr     wp;
  mp_size_t  fn, i;
  mp_limb_t  flimb;

  fn = MRSIZE(m);
  fp = MFP(m);
  wp = RPTR (w);
  xp = RPTR (x);

  CRAY_Pragma ("_CRI ivdep");
  for (i = 0; i < fn; i++)
    {
      flimb = fp[i];
      ADDMOD_1 (wp[i], xp[i], y % flimb, flimb);
    }

  ASSERT_MPR_CRT (w, m);
}

static void
mpr_crt__addmul_ui (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_srcptr  fp, xp;
  mp_size_t  fn, i;
  mp_ptr     wp;
  mp_limb_t  dummy_q, hi, lo, plimb, flimb;

  ASSERT_MPR_CRT (x, m);

  fp = MFP(m);
  fn = MRSIZE(m);
  wp = RPTR (w);
  xp = RPTR (x);

  CRAY_Pragma ("_CRI ivdep");
  for (i = 0; i < fn; i++)
    {
      flimb = fp[i];
      MUL_211 (hi, lo, xp[i], y % flimb);
      UDIV_1121 (dummy_q, plimb, hi, lo, flimb);
      ADDMOD_1 (wp[i], wp[i], plimb, flimb);
    }

  ASSERT_MPR_CRT (w, m);
}

static int
mpr_crt__equal_negone (mpr_srcptr x, mpm_srcptr m)
{
  mp_srcptr  fp, xp;
  mp_size_t  fn, i;

  ASSERT_MPR_CRT (x, m);

  fp = MFP(m);
  fn = MRSIZE(m);
  xp = RPTR (x);
  for (i = 0; i < fn; i++)
    if (xp[i] != fp[i]-1)
      return 0;
  return 1;
}

static int
mpr_crt__equal_one (mpr_srcptr x, mpm_srcptr m)
{
  mp_srcptr  xp;
  mp_size_t  fn, i;

  ASSERT_MPR_CRT (x, m);

  fn = MRSIZE(m);
  xp = RPTR (x);
  for (i = 0; i < fn; i++)
    if (xp[i] != 1)
      return 0;
  return 1;
}

/* ENHANCE-ME: With some pre-calculated inverses should be able to do
   multiplies instead of "%" here. */
static int
mpr_crt__equal_ui (mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_srcptr  xp, fp;
  mp_size_t  fn, i;

  fn = MRSIZE(m);
  fp = MFP(m);
  xp = RPTR (x);

  CRAY_Pragma ("_CRI ivdep");
  for (i = 0; i < fn; i++)
    if (xp[i] != y % fp[i])
      return 0;
  return 1;
}

/* ENHANCE-ME: If fp[i] are primes then can just test for equality.
   ENHANCE-ME: Would want to open-code the gcd to make it vectorize. */
static void
mpr_crt__gcd (mpz_ptr w, mpr_srcptr x, mpm_srcptr m)
{
  mp_srcptr  xp, fp;
  mp_size_t  fn, i;
  mp_limb_t  xlimb, glimb;

  ASSERT (0);

  ASSERT_MPR_CRT (x, m);

  fp = MFP(m);
  fn = MRSIZE(m);
  xp = RPTR (x);

  for (i = 0; i < fn; i++)
    {
      xlimb = xp[i];
      glimb = fp[i];   /* gcd(0,m)==m */
      if (LIKELY (xlimb != 0))
        glimb = mpn_gcd_1 (&fp[i], (mp_size_t) 1, xlimb);

      /* ... */
      ASSERT_ALWAYS (0);
    }

  ASSERT_MPZ (w);
}

/* ENHANCE-ME: This could vectorize if mpn_jacobi_base was inlined, though
   the way that routine needs a data-dependent number of loops might not be
   good.  Could perhaps run it a fixed 2*GMP_NUMB_BITS times or something
   like that.  */
static int
mpr_crt__jacobi (mpr_srcptr x, mpm_srcptr m)
{
  mp_srcptr  fp, xp;
  mp_size_t  fn, i;
  int        result;
  mp_limb_t  xlimb;

  ASSERT_MPR_CRT (x, m);

  if (UNLIKELY ((MPTR(m)[0] & 1) == 0))
    DIVIDE_BY_ZERO;

  fp = MFP(m);
  fn = MRSIZE(m);
  xp = RPTR (x);

  result = 1;
  for (i = 0; i < fn; i++)
    {
      xlimb = xp[i];
      if (UNLIKELY (xlimb == 0))
        return 0;
      result = mpn_jacobi_base (xlimb, fp[i], result);
    }
  return result;
}

static void
mpr_crt__mul (mpr_ptr w, mpr_srcptr x, mpr_srcptr y, mpm_srcptr m)
{
  mp_srcptr  fp, xp, yp;
  mp_size_t  fn, i;
  mp_ptr     wp;
  mp_limb_t  dummy_q, hi, lo;

  ASSERT_MPR_CRT (x, m);
  ASSERT_MPR_CRT (y, m);

  fp = MFP(m);
  fn = MRSIZE(m);
  wp = RPTR (w);
  xp = RPTR (x);
  yp = RPTR (y);

  CRAY_Pragma ("_CRI ivdep");
  for (i = 0; i < fn; i++)
    {
      MUL_211 (hi, lo, xp[i], yp[i]);
      UDIV_1121 (dummy_q, wp[i], hi, lo, fp[i]);
    }

  ASSERT_MPR_CRT (w, m);
}

/* ENHANCE-ME: Use a UDIV_1121 for more than a few bits.
   ENHANCE-ME: Use a powering 2^y mod fp[i] for y > GMP_NUMB_BITS. */
static void
mpr_crt__mul_2exp (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_srcptr  fp, xp;
  mp_size_t  fn, i;
  mp_ptr     wp;
  mp_limb_t  wlimb, xlimb, flimb;

  ASSERT_MPR_CRT (x, m);

  wp = RPTR (w);
  xp = RPTR (x);
  fp = MFP(m);
  fn = MRSIZE(m);

  CRAY_Pragma ("_CRI ivdep");
  for (i = 0; i < fn; i++)
    {
      for (i = 0; i < y; i++)
        {
          xlimb = xp[i];
          flimb = fp[i];
          wlimb = 2 * xlimb;
          wlimb -= ((xlimb & GMP_LIMB_HIGHBIT) != 0 || wlimb >= flimb
                    ? flimb : 0);
          wp[i] = wlimb;
        }
    }

  ASSERT_MPR_CRT (w, m);
}

static void
mpr_crt__mul_ui (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_srcptr  fp, xp;
  mp_size_t  fn, i;
  mp_ptr     wp;
  mp_limb_t  dummy_q, hi, lo, flimb;

  ASSERT_MPR_CRT (x, m);

  fp = MFP(m);
  fn = MRSIZE(m);
  wp = RPTR (w);
  xp = RPTR (x);

  CRAY_Pragma ("_CRI ivdep");
  for (i = 0; i < fn; i++)
    {
      flimb = fp[i];
      MUL_211 (hi, lo, xp[i], y % flimb);
      UDIV_1121 (dummy_q, wp[i], hi, lo, flimb);
    }

  ASSERT_MPR_CRT (w, m);
}

static void
mpr_crt__mul_si (mpr_ptr w, mpr_srcptr x, long y, mpm_srcptr m)
{
  mp_srcptr      fp, xp;
  mp_size_t      fn, i;
  mp_ptr         wp;
  mp_limb_t      dummy_q, hi, lo, flimb, wlimb;
  unsigned long  abs_y;

  ASSERT_MPR_CRT (x, m);

  if (y >= 0)
    {
      mpr_crt__mul_ui (w, x, (unsigned long) y, m);
      return;
    }

  fp = MFP(m);
  fn = MRSIZE(m);
  wp = RPTR (w);
  xp = RPTR (x);
  abs_y = ABS (y);

  CRAY_Pragma ("_CRI ivdep");
  for (i = 0; i < fn; i++)
    {
      flimb = fp[i];
      MUL_211 (hi, lo, xp[i], abs_y % flimb);
      UDIV_1121 (dummy_q, wlimb, hi, lo, flimb);
      NEGMOD_1 (wp[i], wlimb, flimb);
    }

  ASSERT_MPR_CRT (w, m);
}

static void
mpr_crt__neg (mpr_ptr w, mpr_srcptr x, mpm_srcptr m)
{
  mp_srcptr  fp, xp;
  mp_size_t  fn, i;
  mp_ptr     wp;

  ASSERT_MPR_CRT (x, m);

  fp = MFP(m);
  fn = MRSIZE(m);
  wp = RPTR (w);
  xp = RPTR (x);

  CRAY_Pragma ("_CRI ivdep");
  for (i = 0; i < fn; i++)
    NEGMOD_1 (wp[i], xp[i], fp[i]);

  ASSERT_MPR_CRT (w, m);
}

static void
mpr_crt__set (mpr_ptr w, mpr_srcptr x, mpm_srcptr m)
{
  ASSERT_MPR_CRT (x, m);
  MPN_COPY (RPTR(w), RPTR(x), MRSIZE(m));
}

static void
mpr_crt__set_ui (mpr_ptr w, unsigned long x, mpm_srcptr m)
{
  mp_srcptr  fp;
  mp_ptr     wp;
  mp_size_t  fn, i;

  fn = MRSIZE(m);
  fp = MFP(m);
  wp = RPTR (w);

  CRAY_Pragma ("_CRI ivdep");
  for (i = 0; i < fn; i++)
    wp[i] = x % fp[i];
}

static void
mpr_crt__set_si (mpr_ptr w, long x, mpm_srcptr m)
{
  mpr_crt__set_ui (w, ABS(x), m);
  if (x < 0)
    mpr_crt__neg (w, w, m);
}

/* ENHANCE-ME: Perhaps pre-calculate arrays of 2^(i*GMP_NUMB_BITS) mod f[j]
   so mpn_mod_1 can become an inner-product.  Would want a limit on how many
   limbs ("i") were catered for, maybe assume abs(x) < m for most
   applications.  */
static void
mpr_crt__set_z (mpr_ptr w, mpz_srcptr x, mpm_srcptr m)
{
  mp_srcptr  fp, xp;
  mp_ptr     wp;
  mp_size_t  fn, i, xsize, abs_xsize;

  ASSERT_MPZ (x);

  xsize = SIZ(x);
  abs_xsize = ABS(xsize);
  fn = MRSIZE(m);
  fp = MFP(m);
  xp = PTR (x);
  wp = RPTR (w);

  for (i = 0; i < fn; i++)
    wp[i] = mpn_mod_1 (xp, abs_xsize, fp[i]);

  if (xsize < 0)
    mpr_crt__neg (w, w, m);
}

static void
mpr_crt__sub (mpr_ptr w, mpr_srcptr x, mpr_srcptr y, mpm_srcptr m)
{
  mp_srcptr  fp, xp, yp;
  mp_size_t  fn, i;
  mp_ptr     wp;

  ASSERT_MPR_CRT (x, m);
  ASSERT_MPR_CRT (y, m);

  fp = MFP(m);
  fn = MRSIZE(m);
  wp = RPTR (w);
  xp = RPTR (x);
  yp = RPTR (y);

  CRAY_Pragma ("_CRI ivdep");
  for (i = 0; i < fn; i++)
    SUBMOD_1 (wp[i], xp[i], yp[i], fp[i]);

  ASSERT_MPR_CRT (w, m);
}

/* ENHANCE-ME: Pre-calculate some flimb inverses for y%flimb.
   ENHANCE-ME: Note the smallest flimb, so if y < min(flimb) a loop without
   y%flimb can be used (good for small y).  */
static void
mpr_crt__sub_ui (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_srcptr  xp, fp;
  mp_ptr     wp;
  mp_size_t  fn, i;
  mp_limb_t  flimb;

  fn = MRSIZE(m);
  fp = MFP(m);
  wp = RPTR (w);
  xp = RPTR (x);

  CRAY_Pragma ("_CRI ivdep");
  for (i = 0; i < fn; i++)
    {
      flimb = fp[i];
      SUBMOD_1 (wp[i], xp[i], y % flimb, flimb);
    }

  ASSERT_MPR_CRT (w, m);
}

static void
mpr_crt__submul_ui (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_srcptr  fp, xp;
  mp_size_t  fn, i;
  mp_ptr     wp;
  mp_limb_t  dummy_q, hi, lo, plimb, flimb;

  ASSERT_MPR_CRT (x, m);

  fp = MFP(m);
  fn = MRSIZE(m);
  wp = RPTR (w);
  xp = RPTR (x);

  CRAY_Pragma ("_CRI ivdep");
  for (i = 0; i < fn; i++)
    {
      flimb = fp[i];
      MUL_211 (hi, lo, xp[i], y % flimb);
      UDIV_1121 (dummy_q, plimb, hi, lo, flimb);
      SUBMOD_1 (wp[i], wp[i], plimb, flimb);
    }

  ASSERT_MPR_CRT (w, m);
}


/* ENHANCE-ME: Want to take big chunks from _gmp_rand, but need to ensure
   uniformity mod each f[i].  */
void
mpr_crt__urandomm (mpr_ptr w, gmp_randstate_ptr rstate, mpm_srcptr m)
{
  mp_srcptr  fp;
  mp_size_t  fn;
  mp_ptr  wp;
  int     i;

  ASSERT (0);

  fp = MFP(m);
  fn = MRSIZE(m);
  wp = RPTR(w);

  for (i = 0; i < fn; i++)
    {
      _gmp_rand (wp, rstate, GMP_NUMB_BITS * fn);
    }

  ASSERT_MPR_CRT (w, m);
}

void
mpm_init_crt (mpm_ptr m, mp_srcptr fp, mp_size_t fn)
{
  mp_size_t  msize;
  mp_ptr     mp;
  mp_limb_t  c, fl;
  mp_size_t  i;

  if (fn == 0)
    DIVIDE_BY_ZERO;

  MRSIZE(m) = fn;
  MRALLOC(m) = fn;

  ALLOC(MZ(m)) = fn;
  mp = __GMP_ALLOCATE_FUNC_LIMBS (fn);
  PTR(MZ(m)) = mp;

  /* product of fp[] */
  mp[0] = 1;
  msize = 1;
  for (i = 0; i < fn; i++)
    {
      fl = fp[i];
      if (fl < 2 || fl > GMP_NUMB_MAX)
        DIVIDE_BY_ZERO;
      c = mpn_mul_1 (mp, mp, msize, fl);
      mp[msize] = c;
      msize += (c != 0);
    }
  SIZ(MZ(m)) = msize;

#if WANT_ASSERT
  /* pairwise coprime */
  {
    mp_size_t j;
    for (i = 0; i < fn; i++)
      for (j = i+1; j < fn; j++)
        ASSERT (mpn_gcd_1 (&fp[i], (mp_size_t) 1, fp[j]) == 1);
  }
#endif

  MFP(m) = __GMP_ALLOCATE_FUNC_LIMBS (fn);
  MPN_COPY (MFP(m), fp, fn);

  m->_mpr_add          = mpr_crt__add;
  m->_mpr_add_ui       = mpr_crt__add_ui;
  m->_mpr_addmul_ui    = mpr_crt__addmul_ui;
  m->_mpr_clear        = mpr_gen__clear;
  m->_mpr_div          = mpr_gen__div__invert_mul;
  /*   m->_mpr_div_2exp     = mpr_mod__div_2exp; */
  m->_mpr_equal        = mpr_gen__equal;
  m->_mpr_equal_one    = mpr_crt__equal_one;
  m->_mpr_equal_negone = mpr_crt__equal_negone;
  m->_mpr_equal_ui     = mpr_crt__equal_ui;
  m->_mpr_equal_zero   = mpr_gen__equal_zero;
  m->_mpr_gcd          = mpr_crt__gcd;
  /*   m->_mpr_get_z        = mpr_crt__get_z; */
  /*   m->_mpr_invert       = mpr_crt__invert; */
  m->_mpr_jacobi       = mpr_crt__jacobi;
  m->_mpr_mul          = mpr_crt__mul;
  m->_mpr_mul_2exp     = mpr_crt__mul_2exp;
  m->_mpr_mul_si       = mpr_crt__mul_si;
  m->_mpr_mul_ui       = mpr_crt__mul_ui;
  m->_mpr_neg          = mpr_crt__neg;
  m->_mpr_new          = mpr_gen__new;
  /*   m->_mpr_pow          = mpr_crt__pow; */
  m->_mpr_set          = mpr_crt__set;
  m->_mpr_set_z        = mpr_crt__set_z;
  m->_mpr_set_si       = mpr_crt__set_si;
  m->_mpr_set_ui       = mpr_crt__set_ui;
  m->_mpr_sub          = mpr_crt__sub;
  m->_mpr_sub_ui       = mpr_crt__sub_ui;
  m->_mpr_submul_ui    = mpr_crt__submul_ui;
  m->_mpr_swap         = mpr_gen__swap;
  /*   m->_mpr_urandomm     = mpr_crt__urandomm; */
}
